# Copyright 2022 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import pathlib
import typing
import itertools
import copy
import uuid
from simbricks.utils import base as util_base
from simbricks.orchestration.system import base as sys_base
from simbricks.orchestration.system import pcie as sys_pcie
from simbricks.orchestration.system import mem as sys_mem
from simbricks.orchestration.system import eth as sys_eth
from simbricks.orchestration.system.host import disk_images
from simbricks.orchestration.instantiation import (
    socket as inst_socket,
    fragment as inst_fragment,
    dependency_topology as inst_dep_topo,
)
from simbricks.utils import file as util_file

if typing.TYPE_CHECKING:
    from simbricks.orchestration.simulation import base as sim_base
    from simbricks.runtime import command_executor as cmd_exec


class InstantiationEnvironment(util_base.IdObj):

    def __init__(
        self,
        workdir: pathlib.Path = pathlib.Path.cwd() / pathlib.Path("simbricks-workdir"),
        simbricksdir: pathlib.Path = pathlib.Path("/simbricks"),
    ):
        super().__init__()
        self._simbricksdir: str = simbricksdir.resolve()
        self._workdir: str = workdir.resolve()
        self._output_base: str = pathlib.Path(f"{self._workdir}/output").resolve()
        self._tmp_simulation_files: str = pathlib.Path(f"{self._workdir}/tmp").resolve()
        self._imgdir: str = pathlib.Path(f"{self._tmp_simulation_files}/imgs").resolve()
        self._cpdir: str = pathlib.Path(f"{self._tmp_simulation_files}/checkpoints").resolve()
        self._shm_base: str = pathlib.Path(f"{self._tmp_simulation_files}/shm").resolve()


class Instantiation(util_base.IdObj):

    def __init__(
        self,
        sim: sim_base.Simulation,
    ):
        super().__init__()
        self.simulation: sim_base.Simulation = sim
        self.simulation_fragments: set[inst_fragment.Fragment] = set()
        self.fragment_runner_map: dict[inst_fragment.Fragment, str] = dict()
        """Map simulation fragment to runner label."""
        self.runner_label: str | None = None
        """Label of runner we are executing on. Set by runner when fetching
        run."""
        self._simulation_fragment: inst_fragment.Fragment | None = None
        self.env: InstantiationEnvironment | None = None
        self.artifact_name: str = f"simbricks-artifact-{str(uuid.uuid4())}.zip"
        self.artifact_paths: list[str] = []
        self._create_checkpoint: bool = False
        self._restore_checkpoint: bool = False
        self._preserve_checkpoints: bool = True
        self.preserve_tmp_folder: bool = False
        # NOTE: temporary data structure
        self._socket_per_interface: dict[sys_base.Interface, inst_socket.Socket] = {}
        # NOTE: temporary data structure
        self._sim_dependency: inst_dep_topo.SimulationDependencyTopology | None = None
        self._cmd_executor: cmd_exec.CommandExecutorFactory | None = None

    @staticmethod
    def is_absolute_exists(path: str) -> bool:
        path = pathlib.Path(path)
        return path.is_absolute() and path.is_file()

    @property
    def create_artifact(self) -> bool:
        return len(self.artifact_paths) > 0

    @property
    def command_executor(self) -> cmd_exec.CommandExecutorFactory:
        if self._cmd_executor is None:
            raise RuntimeError(f"{type(self).__name__}._cmd_executor should be set")
        return self._cmd_executor

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        if self.runner_label:
            json_obj["runner_label"] = self.runner_label

        json_obj["simulation"] = self.simulation.id()

        fragments_json = []
        if len(self.simulation_fragments) < 1:
            fragment = self.fragment
            util_base.has_attribute(fragment, "toJSON")
            fragments_json.append(fragment.toJSON())
        else:
            for fragment in self.simulation_fragments:
                util_base.has_attribute(fragment, "toJSON")
                fragments_json.append(fragment.toJSON())
        json_obj["simulation_fragments"] = fragments_json

        json_obj["artifact_name"] = self.artifact_name
        json_obj["artifact_paths"] = self.artifact_paths

        json_obj["create_checkpoint"] = self._create_checkpoint
        json_obj["restore_checkpoint"] = self._restore_checkpoint
        json_obj["preserve_checkpoints"] = self._preserve_checkpoints
        json_obj["preserve_tmp_folder"] = self.preserve_tmp_folder

        # TODO: serialize other fields etc. of interest
        return json_obj

    @classmethod
    def fromJSON(cls, sim: sim_base.Simulation, json_obj: dict) -> Instantiation:
        instance = super().fromJSON(json_obj)

        instance.runner_label = util_base.get_json_attr_top_or_none(json_obj, "runner_label")

        simulation_id = int(util_base.get_json_attr_top(json_obj, "simulation"))
        assert simulation_id == sim.id()
        instance.simulation = sim

        instance.simulation_fragments = set()
        fragments_json = util_base.get_json_attr_top(json_obj, "simulation_fragments")
        for frag_json in fragments_json:
            frag_class = util_base.get_cls_by_json(frag_json)
            util_base.has_attribute(frag_class, "fromJSON")
            frag = frag_class.fromJSON(frag_json)
            instance.simulation_fragments.add(frag)

        instance.artifact_name = util_base.get_json_attr_top(json_obj, "artifact_name")
        instance.artifact_paths = util_base.get_json_attr_top(json_obj, "artifact_paths")

        instance._create_checkpoint = bool(
            util_base.get_json_attr_top(json_obj, "create_checkpoint")
        )
        instance._restore_checkpoint = bool(
            util_base.get_json_attr_top(json_obj, "restore_checkpoint")
        )
        instance._preserve_checkpoints = bool(
            util_base.get_json_attr_top(json_obj, "preserve_checkpoints")
        )
        instance.preserve_tmp_folder = bool(
            util_base.get_json_attr_top(json_obj, "preserve_tmp_folder")
        )
        # TODO: deserialize other fields etc. of interest
        return instance

    def _get_opposing_interface(self, interface: sys_base.Interface) -> sys_base.Interface:
        opposing_inf = interface.get_opposing_interface()
        return opposing_inf

    def _opposing_interface_within_same_sim(self, interface: sys_base.Interface) -> bool:
        opposing_interface = self._get_opposing_interface(interface=interface)
        component = interface.component
        opposing_component = opposing_interface.component
        assert interface is not opposing_interface
        return self.find_sim_by_spec(spec=component) == self.find_sim_by_spec(
            spec=opposing_component
        )

    def _updated_tracker_mapping(
        self, interface: sys_base.Interface, socket: inst_socket.Socket
    ) -> None:
        # update interface mapping
        if interface in self._socket_per_interface:
            raise Exception("an interface cannot be associated with two sockets")
        self._socket_per_interface[interface] = socket

    def _get_socket_by_interface(self, interface: sys_base.Interface) -> inst_socket.Socket | None:
        if interface not in self._socket_per_interface:
            return None
        return self._socket_per_interface[interface]

    def _get_opposing_socket_by_interface(
        self, interface: sys_base.Interface
    ) -> inst_socket.Socket | None:
        opposing_interface = self._get_opposing_interface(interface=interface)
        socket = self._get_socket_by_interface(interface=opposing_interface)
        return socket

    def _interface_to_sock_path(self, interface: sys_base.Interface) -> str:
        channel = interface.get_chan_raise()
        queue_ident = f"{channel.a._id}.{channel._id}.{channel.b._id}"

        queue_type = None
        match interface:
            case sys_pcie.PCIeHostInterface() | sys_pcie.PCIeDeviceInterface():
                queue_type = "pci"
            case sys_mem.MemDeviceInterface() | sys_mem.MemHostInterface():
                queue_type = "mem"
            case sys_eth.EthInterface():
                queue_type = "eth"
            case _:
                raise Exception("cannot create socket path for given interface type")

        assert queue_type is not None
        return self._join_paths(
            base=self.shm_base_dir(),
            relative_path=f"{queue_type}-{queue_ident}",
            enforce_existence=False,
        )

    def _create_opposing_socket(
        self, socket: inst_socket.Socket, socket_type: inst_socket.SockType
    ) -> inst_socket.Socket:
        new_ty = (
            inst_socket.SockType.LISTEN
            if socket._type == inst_socket.SockType.CONNECT
            else inst_socket.SockType.CONNECT
        )
        if new_ty != socket_type:
            raise Exception(
                f"cannot create opposing socket, as required type is not supported: required={new_ty.name}, supported={socket_type.name}"
            )
        new_path = socket._path
        new_socket = inst_socket.Socket(path=new_path, ty=new_ty)
        return new_socket

    def update_get_socket(self, interface: sys_base.Interface) -> inst_socket.Socket | None:
        socket = self._get_socket_by_interface(interface=interface)
        if socket:
            return socket
        return None

    def _update_get_socket(
        self, interface: sys_base.Interface, socket_type: inst_socket.SockType
    ) -> inst_socket.Socket:

        if self._opposing_interface_within_same_sim(interface=interface):
            raise Exception(
                "we do not create a socket for channels where both interfaces belong to the same simulator"
            )

        # check if already a socket is associated with this interface
        socket = self._get_socket_by_interface(interface=interface)
        if socket is not None:
            return socket

        # Check if other side already created a socket, and create an opposing one
        socket = self._get_opposing_socket_by_interface(interface=interface)
        if socket is not None:
            new_socket = self._create_opposing_socket(socket=socket, socket_type=socket_type)
            self._updated_tracker_mapping(interface=interface, socket=new_socket)
            return new_socket

        # create socket if opposing socket was not created yet
        sock_path = self._interface_to_sock_path(interface=interface)
        new_socket = inst_socket.Socket(path=sock_path, ty=socket_type)
        self._updated_tracker_mapping(interface=interface, socket=new_socket)
        return new_socket

    def sim_dependencies(self) -> inst_dep_topo.SimulationDependencyTopology:
        if self._sim_dependency is not None:
            return self._sim_dependency
        self._sim_dependency = inst_dep_topo.build_simulation_topology(self)
        return self._sim_dependency

    @property
    def create_checkpoint(self) -> bool:
        """
        Whether to use checkpoint and restore for simulators.

        The most common use-case for this is accelerating host simulator startup
        by first running in a less accurate mode, then checkpointing the system
        state after boot and running simulations from there.
        """
        assert (self._create_checkpoint ^ self._restore_checkpoint) or (
            not self._create_checkpoint and not self._restore_checkpoint
        )
        return self._create_checkpoint

    @create_checkpoint.setter
    def create_checkpoint(self, create_checkpoint: bool) -> None:
        assert (self._create_checkpoint ^ self._restore_checkpoint) or (
            not self._create_checkpoint and not self._restore_checkpoint
        )
        self._create_checkpoint = create_checkpoint

    @property
    def restore_checkpoint(self) -> bool:
        assert (self._create_checkpoint ^ self._restore_checkpoint) or (
            not self._create_checkpoint and not self._restore_checkpoint
        )
        return self._restore_checkpoint

    @restore_checkpoint.setter
    def restore_checkpoint(self, restore_checkpoint: bool) -> None:
        assert (self._create_checkpoint ^ self._restore_checkpoint) or (
            not self._create_checkpoint and not self._restore_checkpoint
        )
        self._restore_checkpoint = restore_checkpoint

    @property
    def fragment(self) -> inst_fragment.Fragment:
        if self._simulation_fragment is not None:
            return self._simulation_fragment

        if self.runner_label is None or not self.simulation_fragments:
            # Experiment does not define any simulation fragments, so
            # implicitly, we create one fragment that spans the whole simulation
            self._simulation_fragment = inst_fragment.Fragment()
            self._simulation_fragment.add_simulators(*self.simulation.all_simulators())
        else:
            fragments = [
                fragment
                for fragment, runner_label in self.fragment_runner_map.items()
                if runner_label == self.runner_label
            ]
            self._simulation_fragment = inst_fragment.Fragment.merged(fragments)
        return self._simulation_fragment

    # TODO: this needs fixing...
    def copy(self) -> Instantiation:
        cop = Instantiation(sim=self.simulation)
        cop.simulation = copy.deepcopy(
            self.simulation
        )  # maybe there is a smarter way of achieving this...
        cop.artifact_name = self.artifact_name
        cop.artifact_paths = self.artifact_paths
        cop._create_checkpoint = self._create_checkpoint
        cop._restore_checkpoint = self._restore_checkpoint
        cop._preserve_checkpoints = self._preserve_checkpoints
        cop.preserve_tmp_folder = self.preserve_tmp_folder
        cop._socket_per_interface = {}
        cop._sim_dependency = None
        return cop

    def out_base_dir(self) -> str:
        return pathlib.Path(f"{self.env._output_base}/{self.simulation.name}/{self._id}").resolve()

    def shm_base_dir(self) -> str:
        return pathlib.Path(f"{self.env._shm_base}/{self.simulation.name}/{self._id}").resolve()

    def imgs_dir(self) -> str:
        return pathlib.Path(f"{self.env._imgdir}/{self.simulation.name}/{self._id}").resolve()

    def cpdir(self) -> str:
        return pathlib.Path(f"{self.env._cpdir}/{self.simulation.name}").resolve()  # /{self._id}"

    def wrkdir(self) -> str:
        return pathlib.Path(f"{self.env._workdir}/{self.simulation.name}").resolve()  # /{self._id}"

    async def prepare(self) -> None:
        to_prepare = [self.shm_base_dir(), self.imgs_dir()]
        if not self.create_checkpoint and not self.restore_checkpoint:
            to_prepare.append(self.cpdir())
        for tp in to_prepare:
            util_file.rmtree(tp)
            util_file.mkdir(tp)

        await self.simulation.prepare(inst=self)

    async def cleanup(self) -> None:
        if self.preserve_tmp_folder:
            return
        to_delete = [self.shm_base_dir(), self.imgs_dir()]
        if not self._preserve_checkpoints:
            to_delete.append(self.cpdir())
        for td in to_delete:
            util_file.rmtree(td)

    def _join_paths(self, base: str = "", relative_path: str = "", enforce_existence=False) -> str:
        if relative_path.startswith("/"):
            raise Exception(
                f"cannot join with base={base} because relative_path={relative_path} starts with '/'"
            )

        joined = pathlib.Path(base).joinpath(relative_path).resolve()
        if enforce_existence and not joined.exists():
            raise Exception(f"couldn't join {base} and {relative_path}")
        return joined.as_posix()

    def join_repo_base(self, relative_path: str) -> str:
        return self._join_paths(
            base=self.env._simbricksdir, relative_path=relative_path, enforce_existence=True
        )

    def join_output_base(self, relative_path: str) -> str:
        return self._join_paths(
            base=self.out_base_dir(),
            relative_path=relative_path,
            enforce_existence=True,
        )

    def hd_path(self, hd_name_or_path: str) -> str:
        if Instantiation.is_absolute_exists(hd_name_or_path):
            return hd_name_or_path
        path = self._join_paths(
            base=self.env._simbricksdir,
            relative_path=f"images/output-{hd_name_or_path}/{hd_name_or_path}",
            enforce_existence=True,
        )
        return path

    def cfgtar_path(self, sim: sim_base.Simulator) -> str:
        return f"{self.wrkdir()}/cfg.{sim.name}.tar"

    # def join_tmp_base(self, relative_path: str) -> str:
    #     return self._join_paths(
    #         base=self.tmp_dir(),
    #         relative_path=relative_path,
    #     )

    def join_imgs_path(self, relative_path: str) -> str:
        return self._join_paths(
            base=self.imgs_dir(),
            relative_path=relative_path,
        )

    def dynamic_img_path(self, img: disk_images.DiskImage, format: str) -> str:
        filename = f"{img._id}.{format}"
        return self._join_paths(
            base=self.imgs_dir(),
            relative_path=filename,
        )

    def hdcopy_path(self, img: disk_images.DiskImage, format: str) -> str:
        filename = f"{img._id}_hdcopy.{format}"
        return self._join_paths(
            base=self.imgs_dir(),
            relative_path=filename,
        )

    def cpdir_subdir(self, sim: sim_base.Simulator) -> str:
        dir_path = f"checkpoint.{sim.full_name()}-{sim._id}"
        return self._join_paths(base=self.cpdir(), relative_path=dir_path, enforce_existence=False)

    def get_simmulator_output_dir(self, sim: sim_base.Simulator) -> str:
        dir_path = f"output.{sim.full_name()}-{sim._id}"
        return self._join_paths(base=self.out_base_dir(), relative_path=dir_path)

    def get_simulator_shm_pool_path(self, sim: sim_base.Simulator) -> str:
        return self._join_paths(
            base=self.shm_base_dir(),
            relative_path=f"{sim.full_name()}-shm-pool-{sim._id}",
        )

    def get_simulation_output_path(self) -> str:
        return self._join_paths(
            base=self.out_base_dir(),
            relative_path=f"out.json",
        )

    def find_sim_by_interface(self, interface: sys_base.Interface) -> sim_base.Simulator:
        return self.find_sim_by_spec(spec=interface.component)

    def find_sim_by_spec(self, spec: sys_base.Component) -> sim_base.Simulator:
        util_base.has_expected_type(spec, sys_base.Component)
        return self.simulation.find_sim(spec)



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import typing as tp
import abc
import io
from simbricks.orchestration.instantiation import base as inst_base
from simbricks.utils import base as utils_base
from simbricks.orchestration.system import base as sys_base

if tp.TYPE_CHECKING:
    from simbricks.orchestration.system import host as sys_host


class Application(utils_base.IdObj):
    def __init__(self, h: sys_host.Host) -> None:
        super().__init__()
        self.host: sys_host.Host = h
        self.wait: bool = False
        self.parameters: dict[tp.Any, tp.Any] = {}

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["host"] = self.host.id()
        json_obj["parameters"] = utils_base.dict_to_json(self.parameters)
        json_obj["wait"] = self.wait
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict):
        instance = super().fromJSON(json_obj)
        host_id = utils_base.get_json_attr_top(json_obj, "host")
        instance.host = system.get_comp(host_id)
        instance.parameters = utils_base.json_to_dict(
            utils_base.get_json_attr_top(json_obj, "parameters")
        )
        instance.wait = bool(utils_base.get_json_attr_top(json_obj, "wait"))
        return instance


# Note AK: Maybe we can factor most of the duplicate calls with the host out
# into a separate module.
class BaseLinuxApplication(Application):
    def __init__(self, h: sys_host.LinuxHost) -> None:
        super().__init__(h)
        self.start_delay: float | None = None
        self.end_delay: float | None = None

    @abc.abstractmethod
    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        """Commands to run on node."""
        raise Exception("must be overwritten")

    def cleanup_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        """Commands to run to cleanup node."""
        if self.end_delay is None:
            return []
        else:
            return [f"sleep {self.start_delay}"]

    def config_files(self, inst: inst_base.Instantiation) -> dict[str, tp.IO]:
        """
        Additional files to put inside the node, which are mounted under
        `/tmp/guest/`.

        Specified in the following format: `filename_inside_node`:
        `IO_handle_of_file`
        """
        return {}

    def prepare_pre_cp(self, inst: inst_base.Instantiation) -> list[str]:
        """Commands to run to prepare node before checkpointing."""
        return []

    def prepare_post_cp(self, inst: inst_base.Instantiation) -> list[str]:
        """Commands to run to prepare node after checkpoint restore."""
        if self.end_delay is None:
            return []
        else:
            return [f"sleep {self.end_delay}"]

    def strfile(self, s: str) -> io.BytesIO:
        """
        Helper function to convert a string to an IO handle for usage in
        `config_files()`.

        Using this, you can create a file with the string as its content on the
        simulated node.
        """
        return io.BytesIO(bytes(s, encoding="UTF-8"))

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["start_delay"] = self.start_delay
        json_obj["end_delay"] = self.end_delay
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict):
        instance = super().fromJSON(system, json_obj)
        instance.start_delay = utils_base.get_json_attr_top_or_none(json_obj, "start_delay")
        instance.end_delay = utils_base.get_json_attr_top_or_none(json_obj, "end_delay")
        return instance


class PingClient(BaseLinuxApplication):
    def __init__(self, h: sys_host.LinuxHost, server_ip: str = "192.168.64.1") -> None:
        super().__init__(h)
        self.server_ip: str = server_ip

    def run_cmds(self, inst: inst_base.Instantiation) -> tp.List[str]:
        return [f"ping {self.server_ip} -c 10"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["server_ip"] = self.server_ip
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict):
        instance = super().fromJSON(system, json_obj)
        instance.server_ip = utils_base.get_json_attr_top(json_obj, "server_ip")
        return instance


class Sleep(BaseLinuxApplication):
    def __init__(self, h: sys_host.LinuxHost, delay: float = 10, infinite: bool = False) -> None:
        super().__init__(h)
        self.infinite: bool = infinite
        self.delay: float = delay

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        if self.infinite:
            return [f"sleep infinity"]
        return [f"sleep {self.delay}"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["infinite"] = self.infinite
        json_obj["delay"] = self.delay
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> Sleep:
        instance = super().fromJSON(system, json_obj)
        instance.infinite = bool(utils_base.get_json_attr_top(json_obj, "infinite"))
        instance.delay = float(utils_base.get_json_attr_top(json_obj, "delay"))
        return instance


class NetperfServer(BaseLinuxApplication):
    def __init__(self, h: sys_host.LinuxHost) -> None:
        super().__init__(h)

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        return ["netserver", "sleep infinity"]

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfServer:
        return super().fromJSON(system, json_obj)


class NetperfClient(BaseLinuxApplication):
    def __init__(self, h: sys_host.LinuxHost, server_ip: str = "192.168.64.1") -> None:
        super().__init__(h)
        self.server_ip: str = server_ip
        self.duration_tp: int = 10
        self.duration_lat: int = 10

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        return [
            "netserver",
            "sleep 0.5",
            f"netperf -H {self.server_ip} -l {self.duration_tp}",
            (
                f"netperf -H {self.server_ip} -l {self.duration_lat} -t TCP_RR"
                " -- -o mean_latency,p50_latency,p90_latency,p99_latency"
            ),
        ]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["server_ip"] = self.server_ip
        json_obj["duration_tp"] = self.duration_tp
        json_obj["duration_lat"] = self.duration_lat
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfClient:
        instance = super().fromJSON(system, json_obj)
        instance.server_ip = utils_base.get_json_attr_top(json_obj, "server_ip")
        instance.duration_tp = int(utils_base.get_json_attr_top(json_obj, "duration_tp"))
        instance.duration_lat = int(utils_base.get_json_attr_top(json_obj, "duration_lat"))
        return instance


class IperfTCPServer(BaseLinuxApplication):

    def __init__(self, h: sys_host.LinuxHost) -> None:
        super().__init__(h)

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        return ["iperf -s -l 32M -w 32M"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfServer:
        return super().fromJSON(system, json_obj)


class IperfUDPServer(BaseLinuxApplication):

    def __init__(self, h: sys_host.LinuxHost) -> None:
        super().__init__(h)

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        return ["iperf -s -u"]

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfServer:
        return super().fromJSON(system, json_obj)


class IperfTCPClient(BaseLinuxApplication):

    def __init__(self, h: sys_host.LinuxHost, server_ip: str = "10.0.0.1", procs: int = 1) -> None:
        super().__init__(h)
        self.server_ip: str = server_ip
        self.procs: int = procs

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        cmds = [f"iperf -l 32M -w 32M  -c {self.server_ip} -i 1 -P {self.procs}"]
        return cmds

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["server_ip"] = self.server_ip
        json_obj["procs"] = self.procs
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfServer:
        instance = super().fromJSON(system, json_obj)
        instance.server_ip = utils_base.get_json_attr_top(json_obj, "server_ip")
        instance.procs = utils_base.get_json_attr_top(json_obj, "procs")
        return instance


class IperfUDPClient(BaseLinuxApplication):

    def __init__(
        self, h: sys_host.LinuxHost, server_ip: str = "10.0.0.1", rate: str = "150m"
    ) -> None:
        super().__init__(h)
        self.server_ip: str = server_ip
        self.rate: str = rate

    def run_cmds(self, inst: inst_base.Instantiation) -> list[str]:
        cmds = [f"iperf -c {self.server_ip} -i 1 -u -b {self.rate}"]
        return cmds

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["server_ip"] = self.server_ip
        json_obj["rate"] = self.rate
        return json_obj

    @classmethod
    def fromJSON(cls, system: sys_base.System, json_obj: dict) -> NetperfServer:
        instance = super().fromJSON(system, json_obj)
        instance.server_ip = utils_base.get_json_attr_top(json_obj, "server_ip")
        instance.rate = utils_base.get_json_attr_top(json_obj, "rate")
        return instance



# Copyright 2024 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import typing as tp
from simbricks.utils import base as util_base

if tp.TYPE_CHECKING:
    from simbricks.orchestration.instantiation import base as inst_base


class System(util_base.IdObj):
    """Defines System configuration of the whole simulation"""

    def __init__(self) -> None:
        super().__init__()
        self._all_components: dict[int, Component] = {}
        self._all_interfaces: dict[int, Interface] = {}
        self._all_channels: dict[int, Channel] = {}

    def _add_component(self, c: Component) -> None:
        assert c.system == self
        assert c.id() not in self._all_components
        self._all_components[c.id()] = c

    def get_comp(self, ident: int) -> Component:
        if ident not in self._all_components:
            raise Exception(f"system object does not store component with id {ident}")

        return self._all_components[ident]

    def _add_interface(self, i: Interface) -> None:
        assert i.component.id() in self._all_components
        assert i.id() not in self._all_interfaces
        self._all_interfaces[i.id()] = i

    def get_inf(self, ident: int) -> Interface:
        if ident not in self._all_interfaces:
            raise Exception(f"system object does not store interface with id {ident}")

        return self._all_interfaces[ident]

    def _add_channel(self, c: Channel) -> None:
        assert c.a.id() in self._all_interfaces and c.b.id() in self._all_interfaces
        assert c.id() not in self._all_channels
        self._all_channels[c.id()] = c

    def get_chan(self, ident: int) -> Channel:
        if ident not in self._all_channels:
            raise Exception(f"system does not store channel with id {ident}")

        return self._all_channels[ident]

    @staticmethod
    def set_latencies(channels: list[Channel], amount: int, ratio: util_base.Time) -> None:
        for chan in channels:
            chan.set_latency(amount, ratio)

    def latencies(
        self, amount: int, ratio: util_base.Time, channel_type: tp.Any | None = None
    ) -> None:
        relevant_channels = self._all_channels
        if channel_type:
            relevant_channels = list(
                filter(
                    lambda chan: util_base.check_type(chan, channel_type),
                    self._all_channels.values(),
                )
            )
        System.set_latencies(relevant_channels, amount, ratio)

    def toJSON(self) -> dict:
        json_obj = super().toJSON()

        components_json = []
        for _, comp in self._all_components.items():
            util_base.has_attribute(comp, "toJSON")
            comp_json = comp.toJSON()
            components_json.append(comp_json)

        json_obj["all_components"] = components_json

        channels_json = []
        for _, chan in self._all_channels.items():
            util_base.has_attribute(chan, "toJSON")
            channels_json.append(chan.toJSON())

        json_obj["channels"] = channels_json

        return json_obj

    @classmethod
    def fromJSON(cls, json_obj: dict) -> System:
        instance = super().fromJSON(json_obj)
        instance._all_components = {}
        instance._all_interfaces = {}
        instance._all_channels = {}

        components_json = util_base.get_json_attr_top(json_obj, "all_components")
        for comp_json in components_json:
            comp_class = util_base.get_cls_by_json(comp_json)
            util_base.has_attribute(comp_class, "fromJSON")
            comp = comp_class.fromJSON(instance, comp_json)

        channels_json = util_base.get_json_attr_top(json_obj, "channels")
        for chan_json in channels_json:
            chan_class = util_base.get_cls_by_json(chan_json)
            util_base.has_attribute(chan_class, "fromJSON")
            chan = chan_class.fromJSON(instance, chan_json)

        return instance


class Component(util_base.IdObj):

    def __init__(self, s: System) -> None:
        super().__init__()
        self.system = s
        self.ifs: list[Interface] = []
        self.parameters: dict[tp.Any, tp.Any] = {}
        s._add_component(self)
        self.name: str | None = None

    def interfaces(self) -> list[Interface]:
        return self.ifs

    # NOTE: overwrite and call in subclasses
    def add_if(self, interface: Interface) -> None:
        self.ifs.append(interface)

    def channels(self) -> list[Channel]:
        return [i.channel for i in self.interfaces() if i.is_connected()]

    async def prepare(self, inst: inst_base.Instantiation) -> None:
        pass

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["system"] = self.system.id()
        json_obj["name"] = self.name
        json_obj["parameters"] = util_base.dict_to_json(self.parameters)

        interfaces_json = []
        for inf in self.interfaces():
            util_base.has_attribute(inf, "toJSON")
            interfaces_json.append(inf.toJSON())
        json_obj["interfaces"] = interfaces_json

        return json_obj

    @classmethod
    def fromJSON(cls, system: System, json_obj: dict) -> Component:
        instance = super().fromJSON(json_obj)
        instance.name = util_base.get_json_attr_top_or_none(json_obj, "name")
        instance.parameters = util_base.json_to_dict(
            util_base.get_json_attr_top(json_obj, "parameters")
        )
        instance.system = system
        system._add_component(instance)

        instance.ifs = []
        interfaces_json = util_base.get_json_attr_top(json_obj, "interfaces")
        for inf_json in interfaces_json:
            inf_class = util_base.get_cls_by_json(inf_json)
            util_base.has_attribute(inf_class, "fromJSON")
            # NOTE: this will add the interface to the system map for retrieval in sub-classes
            inf = inf_class.fromJSON(system, inf_json)

        return instance


class Interface(util_base.IdObj):
    def __init__(self, c: Component) -> None:
        super().__init__()
        self.component = c
        c.system._add_interface(self)
        self.channel: Channel | None = None

    def is_connected(self) -> bool:
        return self.channel is not None

    def disconnect(self) -> None:
        self.channel = None

    def connect(self, c: Channel) -> None:
        if self.channel is not None:
            raise Exception(
                f"cannot connect interface {self.id()} to channel {c.id()}. interface is already connected to channel {self.channel.id()}"
            )
        self.channel = c

    def find_peer(self) -> Interface:
        assert self.channel is not None
        if self.channel.a == self:
            peer_if = self.channel.b
        else:
            peer_if = self.channel.a
        return peer_if

    def get_chan_raise(self) -> Channel:
        if not self.is_connected():
            raise Exception(f"interface(id={self._id}) is not connected to channel")
        return self.channel

    def get_opposing_interface(self) -> Interface:
        chan = self.get_chan_raise()
        return chan.get_opposing_interface(interface=self)

    T = tp.TypeVar("T")

    @staticmethod
    def filter_by_type(interfaces: list[Interface], ty: type[T]) -> list[T]:
        return list(filter(lambda inf: isinstance(inf, ty), interfaces))

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["component"] = self.component.id()
        json_obj["channel"] = self.channel.id()
        return json_obj

    @classmethod
    def fromJSON(cls, system: System, json_obj: dict) -> Interface:
        instance = super().fromJSON(json_obj)
        comp_id = util_base.get_json_attr_top(json_obj, "component")
        comp = system.get_comp(comp_id)
        instance.component = comp
        comp.add_if(instance)
        system._add_interface(instance)
        instance.channel = None
        return instance


class Channel(util_base.IdObj):
    def __init__(self, a: Interface, b: Interface) -> None:
        super().__init__()
        self.latency = 500  # nanoseconds
        self.a: Interface = a
        self.a.connect(self)
        self.b: Interface = b
        self.b.connect(self)
        self.parameters: dict[tp.Any, tp.Any] = {}
        a.component.system._add_channel(self)

    def interfaces(self) -> list[Interface]:
        return [self.a, self.b]

    def disconnect(self):
        # Note AK: this is a bit ugly, this leaves the channel dangling. But
        # it's not referenced anywhere, so that's fine I guess.
        self.a.disconnect()
        self.b.disconnect()

    def get_opposing_interface(self, interface: Interface) -> Interface:
        if interface is not self.a and interface is not self.b:
            raise Exception(
                "cannot determine opposing interface, interface is not connected to channel"
            )
        opposing = self.a if interface is self.b else self.b
        assert opposing != interface
        return opposing

    def set_latency(self, amount: int, ratio: util_base.Time = util_base.Time.Nanoseconds) -> None:
        util_base.has_expected_type(obj=ratio, expected_type=util_base.Time)
        self.latency = amount * ratio

    def toJSON(self) -> dict:
        json_obj = super().toJSON()
        json_obj["latency"] = self.latency
        json_obj["interface_a"] = self.a.id()
        json_obj["interface_b"] = self.b.id()
        json_obj["parameters"] = util_base.dict_to_json(self.parameters)
        return json_obj

    @classmethod
    def fromJSON(cls, system: System, json_obj: dict) -> Channel:
        instance = super().fromJSON(json_obj)
        instance.latency = int(util_base.get_json_attr_top(json_obj, "latency"))
        instance.parameters = util_base.json_to_dict(
            util_base.get_json_attr_top(json_obj, "parameters")
        )

        inf_id_a = int(util_base.get_json_attr_top(json_obj, "interface_a"))
        inf_id_b = int(util_base.get_json_attr_top(json_obj, "interface_b"))
        inf_a = system.get_inf(ident=inf_id_a)
        inf_b = system.get_inf(ident=inf_id_b)
        instance.a = inf_a
        instance.a.connect(instance)
        instance.b = inf_b
        instance.b.connect(instance)

        system._add_channel(instance)

        return instance



# Copyright 2025 Max Planck Institute for Software Systems, and
# National University of Singapore
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Module for building instantiation topology for resolving dependencies between
instantiated components like simulators that determines the order to start them
in."""
from __future__ import annotations

import enum
import typing

from simbricks.orchestration.instantiation import proxy as inst_proxy
from simbricks.orchestration.instantiation import socket as inst_socket
from simbricks.orchestration.simulation import base as sim_base

if typing.TYPE_CHECKING:
    from simbricks.orchestration.instantiation import base as inst_base
    from simbricks.orchestration.system import base as sys_base


class TopologyComponentType(enum.Enum):
    SIMULATOR = enum.auto()
    PROXY = enum.auto()


class TopologyComponent:

    def __init__(
        self,
        type: TopologyComponentType,
        value: sim_base.Simulator | inst_proxy.Proxy,
    ) -> None:
        self.type = type
        self.value = value

    def get_simulator(self) -> sim_base.Simulator:
        if self.type != TopologyComponentType.SIMULATOR:
            raise RuntimeError("Value stored is not a simulator")
        return typing.cast(sim_base.Simulator, self.value)

    def get_proxy(self) -> inst_proxy.Proxy:
        if self.type != TopologyComponentType.PROXY:
            raise RuntimeError("Value stored is not a proxy")
        return typing.cast(inst_proxy.Proxy, self.value)

    def __repr__(self) -> str:
        match self.type:
            case TopologyComponentType.SIMULATOR:
                return str(("sim", self.get_simulator()))
            case TopologyComponentType.PROXY:
                return str(("proxy", self.get_proxy()))
            case _:
                raise RuntimeError("Unhandles type")


SimulationDependencyTopology = typing.NewType(
    "SimulationDependencyTopology",
    dict[TopologyComponent, set[TopologyComponent]],
)
"""Dict mapping from to-be-instantiated component like a simulator to its
dependencies. Dependencies have to start first before component stored as key
can start."""


def build_simulation_topology(
    instantiation: inst_base.Instantiation,
) -> SimulationDependencyTopology:
    # mapping from simulator to its dependencies, i.e., dependencies have to
    # start first before simulator can start
    sim_dependencies: SimulationDependencyTopology = {}

    def insert_dependency(topo_comp_a: TopologyComponent, depends_on: TopologyComponent) -> None:
        if depends_on in sim_dependencies:
            if topo_comp_a in sim_dependencies[depends_on]:
                # TODO: FIXME
                raise Exception("detected cylic dependency, this is currently not supported")

        a_dependencies = set()
        if topo_comp_a in sim_dependencies:
            a_dependencies = sim_dependencies[topo_comp_a]

        a_dependencies.add(depends_on)
        sim_dependencies[topo_comp_a] = a_dependencies

    def update_a_depends_on_b(
        inf_a: sys_base.Interface,
        topo_comp_a: TopologyComponent,
        inf_b: sys_base.Interface,
        topo_comp_b: TopologyComponent,
    ) -> None:
        a_sock = topo_comp_a.value.supported_socket_types(interface=inf_a)
        b_sock = topo_comp_b.value.supported_socket_types(interface=inf_b)

        if a_sock != b_sock:
            if len(a_sock) == 0 or len(b_sock) == 0:
                raise Exception(
                    "cannot create socket and resolve dependency if no socket"
                    " type is supported for an interface"
                )
            if inst_socket.SockType.CONNECT in a_sock:
                assert inst_socket.SockType.LISTEN in b_sock
                insert_dependency(topo_comp_a, depends_on=topo_comp_b)
                instantiation._update_get_socket(
                    interface=inf_a, socket_type=inst_socket.SockType.CONNECT
                )
                instantiation._update_get_socket(
                    interface=inf_b, socket_type=inst_socket.SockType.LISTEN
                )
            else:
                assert inst_socket.SockType.CONNECT in b_sock
                insert_dependency(topo_comp_b, depends_on=topo_comp_a)
                instantiation._update_get_socket(
                    interface=inf_b, socket_type=inst_socket.SockType.CONNECT
                )
                instantiation._update_get_socket(
                    interface=inf_a, socket_type=inst_socket.SockType.LISTEN
                )
        else:
            # deadlock?
            if len(a_sock) != 2 or len(b_sock) != 2:
                raise Exception("cannot solve deadlock")
            # both support both we just pick an order
            insert_dependency(topo_comp_a, depends_on=topo_comp_b)
            instantiation._update_get_socket(
                interface=topo_comp_a, socket_type=inst_socket.SockType.CONNECT
            )
            instantiation._update_get_socket(
                interface=topo_comp_b, socket_type=inst_socket.SockType.LISTEN
            )

    # build dependency graph
    topo_comp_sims: dict[sim_base.Simulator, TopologyComponent] = {}
    topo_comp_proxies: dict[inst_proxy.Proxy, TopologyComponent] = {}
    for sim in instantiation.fragment.all_simulators():
        for comp in sim.components():
            for sim_inf in comp.interfaces():
                if instantiation._opposing_interface_within_same_sim(interface=sim_inf):
                    # both interfaces of channel are located in the same
                    # simulator
                    continue

                topo_comp_a = topo_comp_sims.get(
                    sim, TopologyComponent(TopologyComponentType.SIMULATOR, sim)
                )
                topo_comp_sims[sim] = topo_comp_a
                
                opposing_inf = instantiation._get_opposing_interface(interface=sim_inf)
                if not instantiation.fragment.interface_handled_by_proxy(opposing_inf):
                    sim_b = instantiation.find_sim_by_interface(opposing_inf)
                    topo_comp_b = topo_comp_sims.get(
                        sim_b,
                        TopologyComponent(
                            TopologyComponentType.SIMULATOR,
                            sim_b,
                        ),
                    )
                    topo_comp_sims[sim_b] = topo_comp_b
                else:
                    proxy_b = instantiation.fragment.get_proxy_by_interface(opposing_inf)
                    topo_comp_b = topo_comp_proxies.get(
                        proxy_b,
                        TopologyComponent(
                            TopologyComponentType.PROXY,
                            proxy_b,
                        ),
                    )
                    topo_comp_proxies[proxy_b] = topo_comp_b

                update_a_depends_on_b(sim_inf, topo_comp_a, opposing_inf, topo_comp_b)

    # TODO (Jonas) Add iteration over proxies

    return sim_dependencies
